/* $Id: acl.c,v 1.3 2002/06/23 23:16:04 mickey Exp $ */

#include <vserver.h>
#include <proto.h>
#include <extern.h>

#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <limits.h>

#ifndef WIN32
  #include <unistd.h>
  #include <arpa/inet.h>
#endif

#include "acl.h"

/****************
 *** ACL code ***
 ****************/

/*
 * int acl_add_rule(u_short rule, char action, struct in_addr address,
 *		struct in_addr netmask, time_t expires);
 *
 * Add the rule number <rule> with the action specified by <act> and
 * the <address>, <netmask> and <expires> to the acl database.
 * When the rule was successfully added to the ACL database, the index
 * of the new rule within the ACL table is returned. A value of
 * -1 is returned, if memory allocation/reallocation failed.
 */

int acl_add_rule(u_short rule, char action, struct in_addr address, struct in_addr netmask, time_t expires)
{
	ACLRULE		*newbase;
	ACLRULE		*ruleptr;
	register int	i;

	if(acl_base != NULL && acl_numrules > 0)
	{
		if((newbase = (ACLRULE *)realloc((char *)acl_base, ((acl_numrules + 1) * sizeof(ACLRULE)))) == NULL)
		{
			return -1;
		}

		acl_base = newbase;

		for(i = (acl_numrules - 1); i >= 0; --i)
		{
			ruleptr = RULE(i);

			if(ruleptr->rule > rule)
			{
				newbase = RULE(i+1);
				*newbase = *ruleptr;
				continue;
			}
			else
			{				
				break;
			}

                }

		++i;
		ruleptr = RULE(i);

		ruleptr->rule		= rule;
		ruleptr->action		= action;
		ruleptr->address	= address;
		ruleptr->netmask	= netmask;
		ruleptr->expires	= expires;

		++acl_numrules;

		return i;
	}
	else
	{
		if((acl_base = (ACLRULE *)malloc(sizeof(ACLRULE))) == NULL)
		{
			return -1;
		}

		acl_base->rule		= rule;
		acl_base->action	= action;
		acl_base->address	= address;
		acl_base->netmask	= netmask;
		acl_base->expires	= expires;

		acl_numrules = 1;

		return 0;
	}
}

/*
 * int acl_check_access(struct in_addr address, char *msgbuf);
 *
 * Check the ACL database if the specified <address> is allowed
 * to access. Return 0 if access is granted or -1 if not.
 * If <msgbuf> is a non NULL pointer, the rule granting/denying
 * the access is written to the buffer it points to.
 */

int acl_check_access(struct in_addr address, char *msgbuf)
{
	ACLRULE		*ptr;
	register int	i;

	u_int32_t	tmp1;
	u_int32_t	tmp2;

	for(i = 0; i < acl_numrules; i++)
	{
		ptr = RULE(i);

		tmp1 = address.s_addr;
		tmp2 = ptr->address.s_addr;

		tmp1 &= ptr->netmask.s_addr;
		tmp2 &= ptr->netmask.s_addr;

		/*
		 * Does this rule apply?
		 */

		if((tmp1 == tmp2) || (tmp2 == 0))
		{
			if(msgbuf)
			{
				acl_print_rule(ptr, msgbuf);
			}

			if(ptr->action == 'a')
				return(0);
			else
				return(-1);
		}
	}

	if(msgbuf)
		sprintf(msgbuf, "0xFFFF accept any");

	return(0);
}

/*
 * int acl_delete_rule(u_short rule);
 *
 * Delete the rule number <rule> from the ACL database.
 * Returns a positive index of the previously deleted
 * database record, or a value of 0 if no record with
 * a matching rule number could be found, or the value
 * -1 if a memory reallocation operation failed.
 */

int acl_delete_rule(u_short rule)
{
	ACLRULE		*ptr1;
	ACLRULE		*ptr2;
	register int	i;
	int		idx;

	if((idx = acl_find_rule(rule)) == -1)
		return(0);

	for(i = idx; i < (acl_numrules - 1); i++)
	{
		ptr1 = RULE(i);
		ptr2 = RULE(i + 1);

		*ptr1 = *ptr2;
	}

	if(--acl_numrules == 0)
	{
		free(acl_base);
		acl_base = (ACLRULE *)NULL;
	}
	else
	{
		if((ptr1 = (ACLRULE *)realloc((char *)acl_base, acl_numrules * sizeof(ACLRULE))) == NULL)
		{
			return(-1);		
		}

		acl_base = ptr1;
	}

	return(idx + 1);
}

/*
 * int acl_delete_expired_rules();
 *
 * Delete all ACL rules from the database that have expired.
 */

int acl_delete_expired_rules()
{
	ACLRULE		*ptr;
	register int	i;
	time_t		now;
	int		oldnum = acl_numrules;

	time(&now);

	for(i = acl_numrules - 1; i >= 0; i--)
	{
		ptr = RULE(i);

		if(ptr->expires > 0 && ptr->expires <= now)
		{
			/*
			 * Rule has expired.
			 */

			vlog(VLOG_NOTICE, "ACL rule #%d has expired", ptr->rule);

			if(i < (acl_numrules - 1))
			{
				ACLRULE *src = RULE(i+1);
				BCOPY(src, ptr, sizeof(ACLRULE));
			}

			acl_numrules--;
		}
	}

	/*
	 * Has something changed?
	 */

	if(acl_numrules != oldnum)
	{
		/*
		 * Yes - reallocate buffer to fit new size.
		 */

		if(acl_numrules == 0)
		{
			free(acl_base);
			acl_base = (ACLRULE *)NULL;
		}
		else
		{
			ACLRULE *tp = (ACLRULE *)realloc((char *)acl_base, acl_numrules * sizeof(ACLRULE));

			if(tp == NULL)
				return -1;

			acl_base = tp;		
		}
	}

	return 0;
}

/*
 * int acl_find_rule(u_short rule);
 *
 * Find the first rule in the ACL database that matches rule number <rule>.
 */

int acl_find_rule(u_short rule)
{
	ACLRULE		*ptr;
	register int	i;

	for(i = 0; i < acl_numrules; i++)
	{
		ptr = RULE(i);

		if(ptr->rule == rule)
			return(i);
	}

	return(-1);
}

/*
 * void acl_flush(void);
 *
 * Flush the ACL database.
 */

void acl_flush(void)
{
	if(acl_base != NULL)
	{
		free(acl_base);
		acl_base = (ACLRULE *)NULL;
	}

	acl_numrules = 0;
}

/*
 * int acl_load_file(char *file);
 *
 * Load the ACL database with the contents of the specified <file>.
 * Returns zero if successfull, -1 otherwise.
 */

int acl_load_file(char *file)
{
	acl_flush();
	return(acl_merge_file(file));
}

/*
 * void acl_log_status(void);
 *
 * Log the current ACL database status to the server log
 */

void acl_log_status(void)
{
	if(!acl_numrules)
		vlog(VLOG_INFO, "ACL status: 1 rule active");
	else
		vlog(VLOG_INFO, "ACL status: %d rules active", (acl_numrules + 1));
}

/*
 * int acl_merge_file(char *file);
 *
 * Merge the contents of the specified <file> into the ACL database.
 * Returns zero if successfull, -1 otherwise.
 */

int acl_merge_file(char *file)
{
	char	line[ACL_LINESIZE];
	char	msgbuf[ACL_MSGBUFSIZE];
	char	*s;

	int	cnt = 0;

	FILE	*f;

	/*
	 * Open the specified file for reading
	 */

	if((f = fopen(file, "r")) == NULL)
	{
		vlog(VLOG_ERR, "ERROR: Failed to open ACL configuration file '%s' (%s)",
			file, strerror(errno));
		return(-1);
	}

	vlog(VLOG_INFO, "Reading ACL configuration file '%s'", file);

	/*
	 * Process the file line by line
	 */

	while(fgets((char *)&line[0], ACL_LINESIZE, f) != NULL)
	{
		/*
		 * Make sure the buffer is NUL terminated
		 * and setup line count
		 */

		line[ACL_LINESIZE - 1] = '\0';
		++cnt;

		/*
		 * Filter carriage returns and linefeeds from input
		 */

		while((s = (char *)strrchr(line, '\r')) != NULL)
			*s = '\0';
		while((s = (char *)strrchr(line, '\n')) != NULL)
			*s = '\0';

		/*
		 * Cutoff everything followign a '#' sign
		 */

		while((s = (char *)strrchr(line, '#')) != NULL)
			*s = '\0';

		/*
		 * Ignore empty lines
		 */

		if(line[0] == '\0')
			continue;

		/*
		 * Parse and execute the commandline
		 */

		if(acl_process_cmd(line, msgbuf) == -1)
		{
			vlog(VLOG_ERR, "LINE: %d ERROR: %s\n",
				cnt, msgbuf);
		}
	}

	/*
	 * Close the file
	 */

	fclose(f);

	/*
	 * Done
	 */

	return(0);
}

/*
 * void acl_print_rule(ACLRULE *rule, char *buf);
 *
 * Print the given <rule> in a human readable format
 * into the buffer specified by <buf>.
 */

void acl_print_rule(ACLRULE *rule, char *msgbuf)
{
	int	bits;

	bits = netmask_bits(rule->netmask);

	if(rule->expires > 0)
	{
		time_t	now;
		char *ttl;

		time(&now);
		ttl = pf2time(now, rule->expires);

		if(rule->address.s_addr == INADDR_ANY)
		{
			sprintf(msgbuf, "0x%04X %s any (Expires: %s)", rule->rule,
				(rule->action == 'a') ? "accept" : "reject",
				ttl);
		}
		else if(bits == (CHAR_BIT * sizeof(u_int32_t)))
		{
			sprintf(msgbuf, "0x%04X %s %s (Expires: %s)", rule->rule,
				(rule->action == 'a') ? "accept" : "reject",
				inet_ntoa(rule->address), ttl);
		}
		else
		{
			sprintf(msgbuf, "0x%04X %s %s/%d (Expires: %s)", rule->rule,
				(rule->action == 'a') ? "accept" : "reject",
				inet_ntoa(rule->address), bits, ttl);
		}
	}
	else
	{
		if(rule->address.s_addr == INADDR_ANY)
		{
			sprintf(msgbuf, "0x%04X %s any", rule->rule,
				(rule->action == 'a') ? "accept" : "reject");
		}
		else if(bits == (CHAR_BIT * sizeof(u_int32_t)))
		{
			sprintf(msgbuf, "0x%04X %s %s", rule->rule,
				(rule->action == 'a') ? "accept" : "reject",
				inet_ntoa(rule->address));
		}
		else
		{
			sprintf(msgbuf, "0x%04X %s %s/%d", rule->rule,
				(rule->action == 'a') ? "accept" : "reject",
				inet_ntoa(rule->address), bits);
		}
	}
}

/*
 * int acl_process_cmd(char *cmd, char *msgbuf);
 *
 * Parse the commandline pointed to by <cmd> and
 * process the resulting ACL command. If <msgbuf>
 * is a non NULL pointer, than the result of the
 * operation will be written to the buffer it points
 * to. If the command succeeds, a value of zero is
 * returned, otherwise -1 will be returned.
 */

int acl_process_cmd(char *cmdstr, char *msgbuf)
{
	char		*cmdargv[ACL_MAXARGS];
	char		**aptr;
	char		*mask;
	char		*s;
	char		cmd;
	char		act;

	int		i;
	int		nextarg	= 0;
	int		cmdargc = 0;
	int		mask_bits;

	u_short		rule;

	struct in_addr	address;
	struct in_addr	netmask;

	/*
	 * Split up the line into the argument vector
	 */

#if HAVE_STRSEP
	for(aptr = cmdargv; (*aptr = (char *) strsep(&cmdstr, " \t")) != NULL && cmdargc < ACL_MAXARGS;)
	{
		if(**aptr != '\0')
		{
			++aptr;
			++cmdargc;
		}
	}
#else
	aptr = cmdargv;
	*aptr = (char *)strtok(cmdstr, " \t");

	while(*aptr != NULL && cmdargc < ACL_MAXARGS)
	{
		++aptr;
		++cmdargc;

		*aptr = (char *)strtok(NULL, " \t");
	}
#endif

	/*
	 * No arguments - No action
	 */

	if(cmdargc <= 0)
	{
		if(msgbuf)
			sprintf(msgbuf, "NOP");

		return(0);
	}

	/*
	 * Process the arguments contained in the command vector.
	 * The first argument is either one of the commands "add",
	 * "delete" or "flush". Otherwise the "add" command is
	 * assumed, and the first argument to be a valid rule number.
	 */

	str_lower(cmdargv[0]);

	if(!strcmp(cmdargv[0], "add"))
	{
		cmd = 'a';
		nextarg = 1;
	}
	else if(!strcmp(cmdargv[0], "delete"))
	{
		cmd = 'd';
		nextarg = 1;
	}
	else if(!strcmp(cmdargv[0], "flush"))
	{
		cmd = 'f';
		nextarg = 1;
	}
	else
	{
		cmd = 'a';
		nextarg = 0;

		i = strtol(cmdargv[nextarg], &s, 0);

		if(*cmdargv[nextarg] == '\0' || *s != '\0')
		{
			if(msgbuf)
				sprintf(msgbuf, "syntax error - '%s' unexpected; command or rule number expected", cmdargv[nextarg]);
			return(-1);
		}
	}

	/*
	 * Check for correct number of arguments
	 */

	switch(cmd)
	{
		case 'a':	if((cmdargc - nextarg) > 3)
				{
					if(msgbuf)
						sprintf(msgbuf, "extra argument(s) - add command expects 3 arguments, not %d", (cmdargc - nextarg));
					return(-1);
				}

				if((cmdargc - nextarg) < 3)
				{
					if(msgbuf)
						sprintf(msgbuf, "missing argument(s) - add command expects 3 arguments, not %d", (cmdargc - nextarg));
					return(-1);
				}

				break;

		case 'd':	if((cmdargc - nextarg) > 1)
				{
					if(msgbuf)
						sprintf(msgbuf, "extra argument(s) - delete command expects 1 argument, not %d", (cmdargc - nextarg));
					return(-1);
				}

				if((cmdargc - nextarg) < 1)
				{
					if(msgbuf)
						sprintf(msgbuf, "missing argument - delete command expects an argument");
					return(-1);
				}

				break;

		case 'f':	if((cmdargc - nextarg) > 0)
				{
					if(msgbuf)
						sprintf(msgbuf, "extra argument(s) - flush command does not accept any argument");
					return(-1);
				}

				acl_flush();

				if(msgbuf)
					sprintf(msgbuf, "flush - flushed all rules");

				return(0);
	}

	/*
	 * For the "add" and delete commands, parse the rule number
	 */

	i = strtol(cmdargv[nextarg], &s, 0);

	if(*cmdargv[nextarg] == '\0' || *s != '\0')
	{
		if(msgbuf)
			sprintf(msgbuf, "syntax error - numerical value expected in field %d", nextarg + 1);
		return(-1);		
	}

	if(i < 0 || i > 65534)
	{
		if(msgbuf)
			sprintf(msgbuf, "numerical value for rule number out of range (0 <= n < 65535)");
		return(-1);
	}

	rule = (u_short) i;

	++nextarg;

	/*
	 * Delete command needs only the rule number
	 */

	if(cmd == 'd')
	{
		switch(acl_delete_rule(rule))
		{
			case -1:	if(msgbuf)
						sprintf(msgbuf, "memory reallocation failure");
					return(-1);
			case 0:		if(msgbuf)
						sprintf(msgbuf, "no rule matching #%d", rule);
					return(-1);
			default:	break;
		}

		if(msgbuf)
			sprintf(msgbuf, "delete %d", rule);

		return(0);
	}

	/*
	 * Next argument for an "add" command is the action to be taken
	 */

	str_lower(cmdargv[nextarg]);

	if(!strcmp(cmdargv[nextarg], "allow")
	|| !strcmp(cmdargv[nextarg], "pass")
	|| !strcmp(cmdargv[nextarg], "accept")
	|| !strcmp(cmdargv[nextarg], "grant"))
	{
		act = 'a';	/* Accept */
	}
	else if(!strcmp(cmdargv[nextarg], "deny")
	|| !strcmp(cmdargv[nextarg], "reject")
	|| !strcmp(cmdargv[nextarg], "refuse")
	|| !strcmp(cmdargv[nextarg], "intercept")
	|| !strcmp(cmdargv[nextarg], "void"))
	{
		act = 'r';	/* Reject */
	}
	else
	{
		if(msgbuf)
			sprintf(msgbuf, "syntax error - '%s' unexpeceted, expecting rule action", cmdargv[nextarg]);
		return(-1);
	}

	++nextarg;

	/*
	 * The last argument to the add call is an adress/netmask pair
	 * or the word "any" or "all".
	 */

	str_lower(cmdargv[nextarg]);

	if(!strcmp(cmdargv[nextarg], "any") || !strcmp(cmdargv[nextarg], "all"))
	{
		address.s_addr = INADDR_ANY;
		netmask.s_addr = INADDR_ANY;
		mask_bits = 0;
	}
	else if((mask = (char *)strchr(cmdargv[nextarg], '/')) != NULL)
	{
		*mask++ = '\0';

		if((address.s_addr = inet_addr(cmdargv[nextarg])) == INADDR_NONE)
		{
			if(msgbuf)
				sprintf(msgbuf, "malformed address specified");

			return(-1);
		}

		if((s = (char *)strchr(mask, '.')) != NULL)
		{
			netmask.s_addr = inet_addr(mask);
			mask_bits = netmask_bits(netmask);
		}
		else
		{
			mask_bits = atoi(mask);
			netmask_create(mask_bits, &netmask);
		}
	}
	else
	{
		address.s_addr = inet_addr(cmdargv[nextarg]);
		netmask.s_addr = htonl(0xffffffff);
		mask_bits = netmask_bits(netmask);
	}

	if(acl_add_rule(rule, act, address, netmask, 0) == -1)
	{
		if(msgbuf)
			sprintf(msgbuf, "failed to add rule (memory allocation error)");
		return(-1);
	}

	if(msgbuf)
	{
		if(mask_bits == (CHAR_BIT * sizeof(u_int32_t)))
		{
			sprintf(msgbuf, "add %05d %s %s", rule,
				(act == 'a') ? "accept" : "reject",
				inet_ntoa(address));
		}
		else
		{
			sprintf(msgbuf, "add %05d %s %s/%d", rule,
				(act == 'a') ? "accept" : "reject",
				inet_ntoa(address), mask_bits);
		}
	}

	return(0);
}

/*
 * int netmask_bits(struct in_addr netmask);
 *
 * Get the number of mask bits in the specified netmask
 */

int netmask_bits(struct in_addr netmask)
{
	u_int32_t	mask	= ntohl(netmask.s_addr);
	int	bits	= 0;
	int	flg	= 0;
	int	i;

	for(i = 0; i < (CHAR_BIT * sizeof(u_int32_t)); i++)
	{
		if(mask & 0x00000001)
		{
			flg = 1;
			++bits;
		}
		else
		{
			if(flg)
				return(-1);
		}

		mask = (mask >> 1);
	}

	return(bits);
}

/*
 * void netmask_create(int bits, struct in_addr *netmask);
 *
 * Create a netmask with the number of <bits> specified, and store
 * it into the in_addr struct pointed to by <netmask>.
 */

void netmask_create(int bits, struct in_addr *netmask)
{
	netmask->s_addr = htonl(~0 << ((CHAR_BIT * sizeof(u_int32_t)) - bits));
}
