/* $Id: auth.c,v 1.1 2001/04/09 18:30:43 mickey Exp $ */

#include <vserver.h>
#include <proto.h>
#include <extern.h>
#include <auth.h>

#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>

#ifdef WIN32
  #include <io.h>
#endif

/*
 * int auth_load(void);
 *
 * Load authority database file into memory discarding
 * any previously loaded authority records.
 */

int auth_load(void)
{
	int		fd;
	u_long		records;
	struct stat	st;

	auth_flush();

	/*
	 * Open authority database file
	 */

	if((fd = open(cfdata.auth_file, O_RDWR, 0)) == -1)
	{
		vlog(VLOG_ERR, "ERROR: Unable to open authority file '%s' (%s)",
			cfdata.auth_file, strerror(errno));
		return(-1);
	}

	/*
	 * Determine the number of records in database file,
	 * allocate memory accordingly and load file into memory.
	 */

	if(fstat(fd, &st) == -1)
	{
		vlog(VLOG_ERR, "ERROR: Unable to stat authority file '%s' (%s)",
			cfdata.auth_file, strerror(errno));
		close(fd);
		return(-1);
	}

	if((st.st_size % sizeof(AUTHREC)) != 0 || st.st_size < sizeof(AUTHREC))
	{
		vlog(VLOG_ERR, "ERROR: Authority file '%s' has invalid length",
			cfdata.auth_file);
		close(fd);
		return(-1);
	}

	records = (st.st_size / sizeof(AUTHREC));

	if((auth_base = (AUTHREC *)malloc(st.st_size)) == NULL)
	{
		vlog(VLOG_ERR, "ERROR: Unable to allocate memory (%s)",
			strerror(errno));
		close(fd);
		return(-1);
	}

	if(read(fd, auth_base, st.st_size) != st.st_size)
	{
		vlog(VLOG_ERR, "ERROR: Read error on file '%s' (%s)",
			cfdata.auth_file, strerror(errno));
		close(fd);
		auth_flush();
		return(-1);
	}

	auth_numrec = records;

	close(fd);

	vlog(VLOG_INFO, "%d authority records in database", auth_numrec);

	return(0);
}

/*
 * int auth_save(void);
 *
 * Save all current authority records to the specified file.
 */

int auth_save(void)
{
	int	fd;
#ifdef WIN32
	int	size;
#else
	ssize_t	size;
#endif
	/*
	 * Open database file
	 */

	if((fd = open(cfdata.auth_file, (O_WRONLY | O_TRUNC), 0)) == -1)
	{
		vlog(VLOG_ERR, "ERROR: Failed to open file '%s' for output (%s)",
			cfdata.auth_file, strerror(errno));
		return(-1);
	}

	/*
	 * Write database to file
	 */

	size = (auth_numrec * sizeof(AUTHREC));

	if(write(fd, auth_base, size) != size)
	{
		vlog(VLOG_ERR, "ERROR: Write error on file '%s' (%s)",
			cfdata.auth_file, strerror(errno));
		close(fd);
		return(-1);
	}

	close(fd);

	return(0);
}


#ifndef WIN32

/*
 * int auth_reset(char *rootpasswd);
 *
 * Reset the authority database file to the system defaults.
 */

int auth_reset(char *rootpasswd)
{
	int		fd;
	struct stat	st;
	AUTHREC		auth;

	/*
	 * Check if the root password is valid
	 */

	if(rootpasswd == NULL)
	{
		fprintf(stderr, "ERROR: Invalid root password\n");
		return(-1);
	}

	if(strlen(rootpasswd) < 6)
	{
		fprintf(stderr, "ERROR: Choose a root password with at least 6 characters\n");
		return(-1);
	}

	if(strlen(rootpasswd) > VPPASSWDSIZE)
	{
		fprintf(stderr, "ERROR: Choose a root password with at most %d characters\n", VPPASSWDSIZE);
		return(-1);
	}

	/*
	 * Check if the database file already exists and ask
	 * for confirmation to overwrite the database file.
	 */

	if(stat(cfdata.auth_file, &st) != -1)
	{
		int	i;
		char	c;
		char	auth_backup[512];

		/*
		 * Find next backup version
		 */

		i = 1;

		for(;;)
		{
			sprintf(auth_backup, "%s;%d", cfdata.auth_file, i);

			if(stat(auth_backup, &st) == -1)
				break;

			++i;
		}

		fprintf(stderr, "\nAuthority file '%s' already exists\n\n", cfdata.auth_file);
		fprintf(stderr, "Do you really want to remove the existing authority database file and\n");
		fprintf(stderr, "to create a new one with only the default system managers 'root' account?\n\n");
		fprintf(stderr, "If you answer 'y' then a backup of your existing authority database file\n");
		fprintf(stderr, "will be saved under the name '%s'.\n\n", auth_backup);
		fprintf(stderr, "RESET AUTHORITY DATABASE? ");

		c = getchar();

		if(c != 'y')
		{
			fprintf(stderr, "\nAuthority database file unchanged\n\n");
			return(0);
		}

		if(rename(cfdata.auth_file, auth_backup) == -1)
		{
			fprintf(stderr, "ERROR: Failed to backup file '%s' (%s)\n",
				cfdata.auth_file, strerror(errno));
			return(-1);
		}
	}

	/*
	 * Open database file
	 */

	if((fd = open(cfdata.auth_file, (O_WRONLY | O_TRUNC | O_CREAT), 0600)) == -1)
	{
		fprintf(stderr, "\nERROR: Failed to open file '%s' for output (%s)\n",
			cfdata.auth_file, strerror(errno));
		return(-1);
	}

	/*
	 * Create default system operator account
	 */

	BZERO(&auth, sizeof(AUTHREC));

	sprintf(auth.nick, "%s", "root");
	sprintf(auth.passwd, "%s", rootpasswd);
	time(&auth.account_created);
	time(&auth.last_signon);
	auth.accperm = AUTH_ROOT;
	auth.modperm = AUTH_ROOT;
	auth.sessioncnt = 0;
	auth.cl_opts = DEF_CLOPTMASK;

	/*
	 * Write record to database file
	 */

	if(write(fd, &auth, sizeof(AUTHREC)) != sizeof(AUTHREC))
	{
		fprintf(stderr, "\nERROR: Write error on file '%s' (%s)\n",
			cfdata.auth_file, strerror(errno));
		close(fd);
		return(-1);
	}

	close(fd);

	BZERO(&auth, sizeof(AUTHREC));

	fprintf(stderr, "\nSuccessfully initialized authority file '%s'\n\n",
		cfdata.auth_file);

	return(0);
}
#endif /* !WIN32 */

/*
 * void auth_flush(void);
 *
 * Flush all authority records from the current authority database.
 */

void auth_flush(void)
{
	if(auth_base != NULL)
	{
		free(auth_base);
	}

	auth_numrec = 0;
}

/*
 * int auth_add_record(u_char *nick, u_char *passwd);
 *
 * Add an authority record to the current database.
 */

int auth_add_record(u_char *nick, u_char *passwd)
{
	AUTHREC	*auth;

	if(auth_numrec <= 0)
	{
		if((auth_base = (AUTHREC *)malloc(sizeof(AUTHREC))) == NULL)
		{
			vlog(VLOG_ERR, "ERROR: unable to allocate memory (%s)", strerror(errno));
			return(-1);
		}

		auth = auth_base;
	}
	else
	{
		if((auth = (AUTHREC *)realloc(auth_base, ((auth_numrec + 1) * sizeof(AUTHREC)))) == NULL)
		{
			vlog(VLOG_ERR, "ERROR: unable to reallocate memory (%s)",
				strerror(errno));
			return(-1);
		}

		auth_base = auth;
		auth = GETAUTHREC(auth_numrec);
	}

	BZERO(auth, sizeof(AUTHREC));

	time(&auth->account_created);
	auth->last_signon = auth->last_signoff = 0;
	auth->accperm = auth->modperm = auth->sessioncnt = 0;
	strncpy(auth->nick, nick, VPNICKSIZE);
	strncpy(auth->passwd, passwd, VPPASSWDSIZE);
	auth->cl_opts = DEF_CLOPTMASK;

	++auth_numrec;

	return(0);
}

/*
 * int auth_delete_record(char *nick);
 *
 * Delete the authority record for the specified nickname.
 */

int auth_delete_record(u_char *nick)
{
	register int		i;
	register AUTHREC	*auth;
	register AUTHREC	*prev	= (AUTHREC *)NULL;
	u_char			move	= 0;

	for(i = 0; i < auth_numrec; i++)
	{
		auth = GETAUTHREC(i);

		if(!strncmp(auth->nick, nick, VPNICKSIZE) && !move)
		{
			move = 1;
			prev = auth;
			continue;
		}

		if(move)
		{
			BCOPY(auth, prev, sizeof(AUTHREC));
			prev = auth;
		}
	}

	/*
	 * If move is 0 then the specified nickname was not found,
	 * thus we return error condition (-1);
	 */

	if(!move)
		return(-1);

	/*
	 * Shrink the memory block
	 */

	--auth_numrec;

	if((auth = (AUTHREC *)realloc(auth_base, (auth_numrec * sizeof(AUTHREC)))) == NULL)
	{
		vlog(VLOG_ERR, "ERROR: unable to shrink authority database (%s)",
			strerror(errno));
		return(-1);
	}

	auth_base = auth;

	return(0);
}

/*
 * AUTHREC *auth_getbynick(u_char *nick);
 *
 * Returns a pointer to the authority record for
 * nickname <nick>, or NULL if there is no record
 * associated with the specified nickname.
 */

AUTHREC *auth_getbynick(u_char *nick)
{
	register int		i;
	register AUTHREC	*auth;

	for(i = 0; i < auth_numrec; i++)
	{
		auth = GETAUTHREC(i);

		if(!strncmp(auth->nick, nick, VPNICKSIZE))
			return(auth);
	}

	return((AUTHREC *)NULL);
}
