/* $Id: vclientconn.c,v 1.3 2002/06/23 23:16:07 mickey Exp $ */

#include <vserver.h>
#include <vproto/types.h>
#include <proto.h>
#include <extern.h>

#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>

#ifdef WIN32
  #include <winsock.h>
#else
  #include <sys/socket.h>
#endif

#if HAVE_SYS_UIO_H
  #include <sys/uio.h>
#endif

/****** Code ******/

void assign_client(void)
{
	char			msgbuf[ACL_MSGBUFSIZE];
	char			clienthost[256];
	char			gatewayhost[256];

	u_char			status;
	u_char			lvl;
	char			*cltype;
	char			*serverhost	= NULL;
	u_char			dnschklvl	= DNS_CHK_NONE;
	signed long		chan		= 0;
	u_int32_t		access_perm	= 0;
	u_int32_t		modify_perm	= 0;
	u_int32_t		clflags		= 0;
	u_int32_t		cl_opts		= DEF_CLOPTMASK;
	time_t			last_in		= (time_t)0;

	VIDENT			vid;
	VLOGINDAT		vldat;
	AUTHREC			*auth;

	struct sockaddr_in	local_addr;
	struct sockaddr_in	remote_addr;

	CLDATA *tmp_ptr;

	int			tmp_fd;
	int			local_size	= sizeof(struct sockaddr_in);
	int			remote_size	= sizeof(struct sockaddr_in);

	/*
	 * Accept incoming connection request on original socket and get
	 * the sockaddr structure for each side of the connected peer.
	 */

	if((tmp_fd = accept(sock, NULL, 0)) == -1)
	{
		vlog(VLOG_ERR, "ERROR: accept() returned status %d (%s) - client disconnected",
			errno, strerror(errno));
		return;
	}

	if(GETSOCKNAME(tmp_fd, &local_addr, &local_size) < 0)
	{ 
		vlog(VLOG_ERR, "ERROR: getsockname() returned status %d (%s) - client disconnected",
			errno, strerror(errno));
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		return;
	}

	if(GETPEERNAME(tmp_fd, &remote_addr, &remote_size) < 0)
	{
		vlog(VLOG_ERR, "ERROR: getpeername() returned status %d (%s) - client disconnected",
			errno, strerror(errno));
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		return;
	}

	/*
	 * Check ACL database if the remote address is allowed to connect.
	 */

	if(acl_check_access(remote_addr.sin_addr, msgbuf))
	{
		vlog(VLOG_NOTICE, "Connection rejected HOST: %s:%d IF: %s:%d ACL: %s",
			iptostr(remote_addr.sin_addr), ntohs(remote_addr.sin_port),
			iptostr(local_addr.sin_addr), ntohs(local_addr.sin_port),
			msgbuf);

		vio_write(tmp_fd, ident(CONN_EDENIED), sizeof(VIDENT), 1);
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		return;
	}

	/*
	 * Do a DNS check on the client IP address and get the host name.
	 */

	lvl = dns_check(&remote_addr.sin_addr, clienthost, sizeof(clienthost));

	if(lvl < cfdata.dnschklvl)
	{
		/*
		 * Reject client.
		 */

		vlog(VLOG_NOTICE, "Connection rejected HOST: %s:%d IF: %s:%d REASON: Level %d dns clearance required, level %d too weak.",
			iptostr(remote_addr.sin_addr), ntohs(remote_addr.sin_port),
			iptostr(local_addr.sin_addr), ntohs(local_addr.sin_port),
			cfdata.dnschklvl, lvl);

		vio_write(tmp_fd, ident(CONN_EDNSFAIL), sizeof(VIDENT), 1);
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		return;
	}

	dnschklvl = lvl;

	/*
	 * Get the servername
	 */

	serverhost = iptoname(local_addr.sin_addr);

	vlog(VLOG_INFO, "Connection accepted HOST: %s:%d IF: %s:%d DNSLVL: %d ACL: %s",
		iptostr(remote_addr.sin_addr), ntohs(remote_addr.sin_port),
		iptostr(local_addr.sin_addr), ntohs(local_addr.sin_port),
		dnschklvl, msgbuf);

	/*
	 * Send our VIDENT packet to the client.
	 */

	if(vio_write(tmp_fd, (char *)ident(CONN_OK), sizeof(VIDENT), CONN_TIMEOUT) != sizeof(VIDENT))
	{
		vlog(VLOG_ERR, "ERROR: Failed to send VIDENT packet (%s) - client disconnected", strerror(errno));
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		return;
	}

	/*
	 * Read the remote client's VIDENT packet
	 */

	if(vio_read(tmp_fd, (char *)&vid, sizeof(VIDENT), CONN_TIMEOUT) != sizeof(VIDENT))
	{
		vlog(VLOG_ERR, "ERROR: Failed waiting for client response (%s) - client disconnected", strerror(errno));
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		return;
	}

	vid.ident[VPIDENTSIZE] = '\0';

	/*
	 * Check the magic number supplied with the client's VIDENT packet.
	 */

	if(ntohl(vid.magic) != VCLIENT_MAGIC)
	{
		vlog(VLOG_ERR, "ERROR: Magic number mismatch ($%08x/$%08x) - client disconnected",
			ntohl(vid.magic), VCLIENT_MAGIC);
		status = CONN_EMAGIC;
		goto errstatus;
	}

	/*
	 * Check if the client's protocol revision is compatible with ours.
	 */

	if(MAJOR_REV(ntohl(vid.proto)) != MAJOR_REV(PROTO_REV))
	{
		vlog(VLOG_ERR, "ERROR: Invalid protocol revision - client disconnected");
		status = CONN_EPROTO;
		goto errstatus;
	}

	/*
	 * Get the initial channel number from the high order 24 bit of the
	 * type field in the client's VIDENT packet. This is an extension to
	 * the original VChat protocol.
	 */

	if(ntohl(vid.type) & 0x80000000)
		chan = (((ntohl(vid.type) & 0xffffff00) >> 8) | 0xff000000);
	else
		chan = ((ntohl(vid.type) & 0xffffff00) >> 8);

	vid.type = htonl(ntohl(vid.type) & 0x000000ff);

	/*
	 * Check if the supplied client type is valid
	 */

	switch(ntohl(vid.type))
	{
		case CLIENT_FRONTEND:	cltype = "frontend";
					break;

		case CLIENT_GATEWAY:	cltype = "gateway";
					break;

		default:		vlog(VLOG_ERR, "ERROR: Client type $%x invalid - client disconnected", ntohl(vid.type));
					status = CONN_ECLTYPE;
					goto errstatus;
	}

	vlog(VLOG_INFO, "Remote client IDENT: %s TYPE: %s PROTOCOL: %x-%x",
		vid.ident, cltype, MAJOR_REV(ntohl(vid.proto)),
		MINOR_REV(ntohl(vid.proto)));

	/*
	 * Client ID accepted - signal OK to the client.
	 */

	status = CONN_OK;

	if(vio_write(tmp_fd, &status, 1, CONN_TIMEOUT) != 1)
	{
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		vlog(VLOG_ERR, "ERROR: Failed to send status (%s) - client disconnected", strerror(errno));

		return;
	}

	/*
	 * Read the client's VLOGINDAT packet.
	 */

	if(vio_read(tmp_fd, (char *)&vldat, sizeof(VLOGINDAT), CONN_TIMEOUT) == -1)
	{
#ifdef WIN32
		closesocket(tmp_fd);
#else
		close(tmp_fd);
#endif
		vlog(VLOG_ERR, "ERROR: Failed to read VLOGINDAT (%s) - client disconnected", strerror(errno));

		return;
	}

	vldat.uname[VPUSERSIZE] = vldat.nick[VPNICKSIZE] = '\0';

	/*
	 * For clients of type CLIENT_GATEWAY, check ACL database if
	 * the apparent client address is allowed to access.
	 */

	if(ntohl(vid.type) == CLIENT_GATEWAY)
	{
		/*
		 * The address of the connecting party is actually the
		 * gateway address.
		 */

		strncpy(gatewayhost, clienthost, sizeof(gatewayhost));
		BZERO(clienthost, sizeof(clienthost));

		/*
		 * Check the ACLs if the actual client address is allowed.
		 */

		if(acl_check_access(vldat.remote_user_ip, msgbuf))
		{
			vlog(VLOG_NOTICE, "Gateway user rejected HOST: %s VIA: %s ACL: %s",
				iptostr(vldat.remote_user_ip),
				iptostr(remote_addr.sin_addr), msgbuf);

			status = CONN_EDENIED;
			goto errstatus;
		}

		/*
		 * Check DNS for the apparent client address.
		 */

		lvl = dns_check(&vldat.remote_user_ip, clienthost, sizeof(clienthost));

		if(lvl < cfdata.dnschklvl)
		{
			/*
			 * Reject client.
			 */

			vlog(VLOG_NOTICE, "Gateway user rejected HOST: %s VIA: %s REASON: Level %d dns clearance required, level %d too weak.",
				iptostr(vldat.remote_user_ip),
				iptostr(remote_addr.sin_addr),
				cfdata.dnschklvl, lvl);

			status = CONN_EDNSFAIL;
			goto errstatus;
		}

		/*
		 * The DNS check level for this client will be the lower
		 * of the gateway address check level and the client address
		 * check level.
		 */

		if(lvl < dnschklvl)
			dnschklvl = lvl;

		/*
		 * Log the news.
		 */

		vlog(VLOG_INFO, "Gateway user accepted HOST: %s VIA: %s DNSLVL: %d ACL: %s",
			iptostr(vldat.remote_user_ip),
			iptostr(remote_addr.sin_addr),
			dnschklvl, msgbuf);

		/*
		 * Connections of type GATEWAY do not
		 * have a username.
		 */

		sprintf(vldat.uname, "-none-");
	}

	/*
	 * Empty nicknames are translated to
	 * be equal to the username.
	 */

	if(!strlen(vldat.nick))
		strcpy(vldat.nick, vldat.uname);

	/*
	 * Reject the client if the supplied nickname is empty or too large
	 */

	if(vldat.nick[0] == '\0' || strlen(vldat.nick) > VPNICKSIZE)
	{
		vlog(VLOG_NOTICE, "Nickname `%s' too long - client rejected", vldat.nick);
		status = CONN_ENSIZE;
		goto errstatus;
	}

	/*
	 * Reject client if the supplied nickname contains invalid chars
	 */

	if(!nick_ok(vldat.nick))
	{
		vlog(VLOG_NOTICE, "Nickname '%s' invalid - client rejected", vldat.nick);
		status = CONN_ENINVAL;
		goto errstatus;
	}

	/*
	 * Check if there is an authority record
	 * matching the supplied nickname
	 */

	if((auth = auth_getbynick(vldat.nick)) == NULL)
	{
		/*
		 * No authority record for supplied nickname found
		 * If the server is in closed mode, then reject any
		 * unregistered nicknames.
		 */

		if(cfdata.flags & CF_REG_ONLY)
		{
			vlog(VLOG_NOTICE, "Nickname '%s' not registered - client rejected", vldat.nick);
			status = CONN_EDENIED;
			goto errstatus;
		}

		/*
		 * Check for duplicated nicknames
		 */

		if(nick_dup(vldat.nick))
		{
			vlog(VLOG_NOTICE, "Duplicated nickname '%s' invalid - client rejected", vldat.nick);
			status = CONN_ENDUP;
			goto errstatus;
		}
	}
	else
	{
		VPASSWD	vpwd;

		/*
		 * Authority record found - check client protocol revision.
		 */

		if(MINOR_REV(ntohl(vid.proto)) < 0x0e)
		{
			/*
			 * Protocol Revision 4-d or below
			 *
			 * COMPAT: Return CONN_ENDUP to let the
			 *         client think the nickname is
			 *         already assigned.
			 */

			vlog(VLOG_NOTICE, "Supplied nickname '%s' needs authorization - client rejected", vldat.nick);
			status = CONN_ENDUP;
			goto errstatus;
		}

		/*
		 * Protocol Revision 4-e or above - request password
		 */

		status = CONN_AUTHREQ;

		if(vio_write(tmp_fd, &status, 1, CONN_TIMEOUT) == -1)
		{
#ifdef WIN32
			closesocket(tmp_fd);
#else
			close(tmp_fd);
#endif
			vlog(VLOG_ERR, "ERROR: Failed to send status (%s) - client disconnected",
				strerror(errno));

			return;
		}

		/*
		 * Read authorization
		 */

		BZERO(&vpwd, sizeof(VPASSWD));

		if(vio_read(tmp_fd, &vpwd, sizeof(VPASSWD), CONN_TIMEOUT) == -1)
		{
#ifdef WIN32
			closesocket(tmp_fd);
#else
			close(tmp_fd);
#endif
			vlog(VLOG_ERR, "ERROR: Failed to read VPASSWD (%s) - client disconnected", strerror(errno));
			return;
		}

		vpwd.passwd[VPPASSWDSIZE] = '\0';

		/*
		 * Check the supplied password against authority record
		 */

		if(strncmp(vpwd.passwd, auth->passwd, VPPASSWDSIZE))
		{
			/*
			 * Password does not match!
			 */

			vlog(VLOG_NOTICE, "Invalid password for nickname '%s' - client rejected", vldat.nick);
			status = CONN_EAUTHFAIL;
			goto errstatus;
		}

		/*
		 * Password matched, set permissions,
		 * update last login and session counter
		 */

		vlog(VLOG_INFO, "Password ok for nickname '%s'",
			vldat.nick);

		access_perm = auth->accperm;
		modify_perm = auth->modperm;

		clflags |= CFLAG_REGISTERED;

		last_in = auth->last_signon;
		time(&auth->last_signon);
		auth->last_signoff = 0xffffffff;
		++auth->sessioncnt;

		/*
		 * Get saved client options from authority record.
		 */

		cl_opts = auth->cl_opts;

		/*
		 * Update authority record.
		 */

		auth_save();
	}

	/*
	 * Allocate space for a new CLDATA client record in the database.
	 */

	if(cl_buf)
	{
		if((tmp_ptr = (CLDATA *)realloc(cl_buf, ((clients + 1) * sizeof(CLDATA)))) == (CLDATA *)0)
		{
			vlog(VLOG_ERR, "ERROR: memory reallocation failure (%s) - client rejected", strerror(errno));
			status = CONN_ENOMEM;
			goto errstatus;
		}

		cl_buf = tmp_ptr;
	}
	else
	{
		if((cl_buf = (CLDATA *)malloc(sizeof(CLDATA))) == (CLDATA *)0)
		{
			vlog(VLOG_ERR, "ERROR: memory allocation failure (%s) - client rejected", strerror(errno));
			status = CONN_ENOMEM;
			goto errstatus;
		}
	}

	tmp_ptr = cl_addr(clients);
	BZERO(tmp_ptr, sizeof(CLDATA));

	/*
	 * Fill in the client data into the CLDATA record.
	 */

	tmp_ptr->fd = tmp_fd;
	tmp_ptr->type = ntohl(vid.type);
	tmp_ptr->flags = clflags;
	tmp_ptr->error = 0;
	tmp_ptr->accperm = access_perm;
	tmp_ptr->modperm = modify_perm;
	tmp_ptr->channel = chan;
	tmp_ptr->pubmsgcnt = tmp_ptr->prvmsgcnt = 0;
	tmp_ptr->protocol = ntohl(vid.proto);

	time(&tmp_ptr->login);
	timestamp(&tmp_ptr->lastsend);
	tmp_ptr->stamp = tmp_ptr->login;

	strcpy(tmp_ptr->user, vldat.uname);
 	strcpy(tmp_ptr->nick, vldat.nick);
	strcpy(tmp_ptr->ident, vid.ident);

	tmp_ptr->if_addr = local_addr.sin_addr;

	BZERO(&tmp_ptr->if_name, sizeof(tmp_ptr->if_name));

	if(cfdata.servername[0] != '\0')
		strncpy(tmp_ptr->if_name, cfdata.servername, sizeof(tmp_ptr->if_name)-1);
	else
		strncpy(tmp_ptr->if_name, serverhost, sizeof(tmp_ptr->if_name)-1);

	BZERO(&tmp_ptr->host, sizeof(tmp_ptr->host));
	BZERO(&tmp_ptr->gw_name, sizeof(tmp_ptr->gw_name));

	if(ntohl(vid.type) == CLIENT_GATEWAY)
	{
		tmp_ptr->gw_addr = remote_addr.sin_addr;
		tmp_ptr->ip_addr = vldat.remote_user_ip;
		strncpy(tmp_ptr->host, clienthost, sizeof(tmp_ptr->host) - 1);
		strncpy(tmp_ptr->gw_name, gatewayhost, sizeof(tmp_ptr->gw_name) - 1);
	}
	else
	{
		tmp_ptr->ip_addr = remote_addr.sin_addr;
		strncpy(tmp_ptr->host, clienthost, sizeof(tmp_ptr->host) - 1);
	}

	tmp_ptr->cl_opts = cl_opts;
	tmp_ptr->dnschklvl = dnschklvl;

	++clients;
	++vs.vs_total_conn;
	++vs.vs_active_conn;

	/*
	 * Signal OK to the client and sign it on
	 */

	status = CONN_OK;

	if(vio_write(tmp_fd, &status, 1, CONN_TIMEOUT) == -1)
	{
		vlog(VLOG_ERR, "ERROR: Failed to send status (%s) - client disconnected", strerror(errno));
		delete_client(tmp_fd);
	}
	else
	{
		/*
		 * Signon the new client to the conference
		 */

		++vs.vs_requests_signon;
		++vs.vs_requests_processed;
		cmd_signon(tmp_ptr, last_in);
	}

	return;

errstatus:

	vio_write(tmp_fd, &status, 1, 1);
#ifdef WIN32
	closesocket(tmp_fd);
#else
	close(tmp_fd);
#endif
	return;
}

void delete_client(int cl_fd)
{
	int cl_num;

	if((cl_num = getclientnum(cl_fd)) == -1)
	{
		vlog(VLOG_ERR, "ERROR: Unable to delete client ID %d - not in database", cl_fd);
		return;
	}

	if(clients > 1)
	{
		if(cl_num < (clients - 1))
		{
			BCOPY(cl_addr((cl_num + 1)), cl_addr(cl_num), ((clients - (cl_num + 1)) * sizeof(CLDATA)));
		}

		--clients;

		if((cl_buf = (CLDATA *)realloc(cl_buf, (clients * sizeof(CLDATA)))) == (CLDATA *)0)
			vlog(VLOG_ERR, "ERROR: Memory reallocation failed (%s)", strerror(errno));
	}
	else
	{
		free(cl_buf);
		cl_buf = NULL;
		clients = 0;
	}

#ifdef WIN32
	closesocket(cl_fd);
#else
	close(cl_fd);
#endif
	vs.vs_active_conn--;
}

int getclientnum(int fd)
{
	int i;

	CLDATA *ptr;

	for(i = 0; i < clients; i++)
	{
		ptr = cl_addr(i);

		if(ptr->fd == fd)
			return(i);
	}

	return(-1);
}

void snd_svmsg(CLDATA *cldat, char *fmt, ...)
{
	va_list xp;


#if HAVE_SYS_UIO_H
	char		sndbuf[SVMSGBUF];
	struct iovec	iov[2];
	VMSG		svmsg;
#else
	char		sndbuf[SVMSGBUF + sizeof(VMSG)];
	char		*datbuf;
	VMSG		*svmsg;
#endif

	va_start(xp, fmt);

	if(cldat->flags & CFLAG_ERROR)
	{
		va_end(xp);
		return;
	}

#if HAVE_SYS_UIO_H
	vsprintf(sndbuf, fmt, xp);

	svmsg.cmd = htonl(SERV_MSG);
	svmsg.len = htonl(strlen(sndbuf));

	iov[0].iov_base = (char *)&svmsg;
	iov[1].iov_base = sndbuf;
	iov[0].iov_len = sizeof(VMSG);
	iov[1].iov_len = strlen(sndbuf);

	if(vio_writev(cldat->fd, (struct iovec *)&iov, 2, CONN_TIMEOUT) == -1)
#else
	svmsg = (VMSG *)&sndbuf;
	datbuf = (char *)&sndbuf[sizeof(VMSG)];

	vsprintf(datbuf, fmt, xp);

	svmsg->cmd = htonl(SERV_MSG);
	svmsg->len = htonl(strlen(datbuf));

	if(vio_write(cldat->fd, (char *)&sndbuf, ntohl(svmsg->len) + sizeof(VMSG), CONN_TIMEOUT) == -1)
#endif
	{
		cldat->flags |= CFLAG_ERROR;
		cldat->error = errno;
	}

	va_end(xp);
}

void snd2client(CLDATA *cldat, u_short packtype, char *data, size_t datalen)
{
#if HAVE_SYS_UIO_H
	struct iovec	iov[2];
	VMSG		svmsg;

	if(cldat->flags & CFLAG_ERROR)
		return;

	svmsg.cmd = htonl(packtype);
	svmsg.len = (data && datalen) ? htonl(datalen) : 0;

	iov[0].iov_base = (char *)&svmsg;
	iov[0].iov_len = sizeof(VMSG);

	if(data && datalen)
	{
		iov[1].iov_base = data;
		iov[1].iov_len = datalen;

		if(vio_writev(cldat->fd, (struct iovec *)&iov, 2, CONN_TIMEOUT) == -1)
		{
			cldat->flags |= CFLAG_ERROR;
			cldat->error = errno;
		}
	}
	else
	{
		if(vio_writev(cldat->fd, (struct iovec *)&iov, 1, CONN_TIMEOUT) == -1)
		{
			cldat->flags |= CFLAG_ERROR;
			cldat->error = errno;
		}
	}

#else	/* !HAVE_SYS_UIO_H */

	int		bufsize;
	char		*sndbuf;
	char		*datbuf;
	VMSG		*svmsg;

	if(cldat->flags & CFLAG_ERROR)
		return;

	bufsize = (data && datalen) ? (sizeof(VMSG) + datalen) : sizeof(VMSG);

	if((sndbuf = malloc(bufsize)) == NULL)
	{
		vlog(VLOG_ERR, "ERROR: Memory allocation failed (%s)",
			strerror(errno));

		cldat->flags |= CFLAG_ERROR;
		cldat->error = errno;

		return;
	}

	svmsg = (VMSG *)sndbuf;

	if(data && datalen)
	{
		datbuf = (sndbuf + sizeof(VMSG));
		BCOPY(data, datbuf, datalen);
	}
	else
	{
		datbuf = NULL;
	}

	svmsg->cmd = htonl(packtype);
	svmsg->len = (data && datalen) ? htonl(datalen) : 0;

	if(vio_write(cldat->fd, sndbuf, bufsize, CONN_TIMEOUT) == -1)
	{
		cldat->flags |= CFLAG_ERROR;
		cldat->error = errno;
	}

	free(sndbuf);

#endif	/* HAVE_SYS_UIO_H */
}
