/* $Id: vio.c,v 1.1 2001/04/09 18:30:45 mickey Exp $ */

#include <vserver.h>
#include <proto.h>
#include <extern.h>

#ifndef WIN32	/* ----------------------- UN*X ---------------------------- */

#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <unistd.h>

#if HAVE_SYS_UIO_H
  #include <sys/uio.h>
#endif

/*** Globals ***/

jmp_buf timeout;

/*
 * --> void conn_timeout(void) <--
 *
 * Timeout function called by SIGALRM signal upon expiry of
 * a read or write operation timeout period. Uses longjmp to
 * return to the last setjmp(2).
 */

void conn_timeout(void)
{
	alarm(0);
	longjmp(timeout, 1);
}

/*
 * --> ssize_t vio_write(int fd, void *buf, size_t nbytes, time_t tout); <--
 *
 * Attempt to write <nbytes> of data from the buffer pointed to by <buf>
 * to the socket descriptor <fd>. If the data cannot be written within
 * <tout> seconds, this function returns -1 with errno set to ETIMEDOUT.
 */

ssize_t vio_write(int fd, void *buf, size_t nbytes, time_t tout)
{
#if HAVE_SIGACTION
	struct sigaction sa;

	sa.sa_handler = (void *)conn_timeout;
	sigemptyset(&sa.sa_mask);

  #ifdef SA_INTERRUPT
	sa.sa_flags = SA_INTERRUPT;
  #else
	sa.sa_flags = 0;
  #endif

	sigaction(SIGALRM, &sa, NULL);
#else
	struct sigvec sv = {(void *)conn_timeout, 0, SV_INTERRUPT};

	sigvec(SIGALRM, &sv, NULL);
#endif

	if(!setjmp(timeout))
	{
		ssize_t ret;
		ssize_t sent = 0;

		alarm(tout);

		while(sent < nbytes)
		{
			if((ret = write(fd, buf, (nbytes - sent))) >= 0)
			{
				buf = (u_char *)buf + ret;
				sent += ret;
			}
			else
			{
				alarm(0);
				return(ret);
			}
		}

		alarm(0);
		return(sent);
	}
	else
	{
		sigset_t smask;

		sigemptyset(&smask);
		sigaddset(&smask, SIGALRM);
		sigprocmask(SIG_UNBLOCK, &smask, NULL);

		errno = ETIMEDOUT;
		vlog(VLOG_ERR, "ERROR: Output operation timed out");
		return(-1);
	}
}

#if HAVE_SYS_UIO_H

/*
 * --> ssize_t vio_writev(int fd, struct iovec *iov, int iovcnt, time_t tout); <--
 *
 * Attempt to write the contents of the buffers pointed to by the iovec array
 * <iov> whose size is determined by <iovcnt> to the socket descriptor <fd>.
 * If the write operation does not finish within <tout> seconds this
 * function returns -1 with errno set to ETIMEDOUT.
 */

ssize_t vio_writev(int fd, struct iovec *iov, int iovcnt, time_t tout)
{
#if HAVE_SIGACTION
	struct sigaction sa;

	sa.sa_handler = (void *)conn_timeout;
	sigemptyset(&sa.sa_mask);

  #ifdef SA_INTERRUPT
	sa.sa_flags = SA_INTERRUPT;
  #else
	sa.sa_flags = 0;
  #endif

	sigaction(SIGALRM, &sa, NULL);
#else
	struct sigvec sv = {(void *)conn_timeout, 0, SV_INTERRUPT};

	sigvec(SIGALRM, &sv, NULL);
#endif

	if(!setjmp(timeout))
	{
		ssize_t x;

		alarm(tout);

		x = writev(fd, iov, iovcnt);

		alarm(0);
		return((x > 0) ? x : -1);
	}
	else
	{
		sigset_t smask;

		sigemptyset(&smask);
		sigaddset(&smask, SIGALRM);
		sigprocmask(SIG_UNBLOCK, &smask, NULL);

		errno = ETIMEDOUT;
		vlog(VLOG_ERR, "ERROR: Output operation timed out");
		return(-1);
	}
}

#endif /* HAVE_SYS_UIO_H */

/*
 * --> ssize_t vio_read(int fd, void *buf, size_t nbytes, time_t tout); <--
 *
 * Attempt to read <nbytes> from socket descriptor <fd> into the buffer
 * pointed to by <buf>. If the corresponding number of bytes cannot be
 * read withing <tout> seconds, the function returns -1 with errno set
 * to ETIMEDOUT.
 */

ssize_t vio_read(int fd, void *buf, size_t nbytes, time_t tout)
{
#if HAVE_SIGACTION
	struct sigaction sa;

	sa.sa_handler = (void *)conn_timeout;
	sigemptyset(&sa.sa_mask);

  #ifdef SA_INTERRUPT
	sa.sa_flags = SA_INTERRUPT;
  #else
	sa.sa_flags = 0;
  #endif

	sigaction(SIGALRM, &sa, NULL);
#else
	struct sigvec sv = {(void *)conn_timeout, 0, SV_INTERRUPT};

	sigvec(SIGALRM, &sv, NULL);
#endif

	if(!setjmp(timeout))
	{
		ssize_t ret;
		ssize_t got = 0;

		alarm(tout);

		while(got < nbytes)
		{
			switch(ret = read(fd, buf, (nbytes - got)))
			{
				case 0:		errno = EPIPE;
				case -1:	alarm(0);
						return(-1);
				default:	got += ret;
						buf = (u_char *)buf + ret;
						break;
			}
		}

		alarm(0);

		return(got);
	}
	else
	{
		sigset_t smask;

		sigemptyset(&smask);
		sigaddset(&smask, SIGALRM);
		sigprocmask(SIG_UNBLOCK, &smask, NULL);

		errno = ETIMEDOUT;
		vlog(VLOG_ERR, "ERROR: Input operation timed out");
		return(-1);
	}
}



#else /* ------------------------------ WIN32 ------------------------------ */

#include <winsock.h>


/*************************************
 * WINDOWS 32 low level I/O routines *
 *************************************/

DWORD WINAPI WriteOp(LPVOID arg)
{
	IO_OP	*op	= (IO_OP *)arg;
	DWORD	sent	= 0;
	void	*buf	= op->buf;
	int	ret;

	while(sent < op->len)
	{
		if((ret = send(op->fd, buf, (op->len - sent), 0)) != SOCKET_ERROR)
		{
			buf = (u_char *)buf + ret;
			sent += ret;
		}
		else
		{
			ExitThread(-1);
		}
	}

	return(sent);
}

ssize_t vio_write(int fd, void *buf, size_t nbytes, time_t tout)
{
	IO_OP	op;
	HANDLE	hOutputThread		= NULL;
	DWORD	dwOutputThreadID	= 0;
	DWORD	dwRetCode		= -1;

	/*
	 * Create a Thread that will perform the requested output operation.
	 */

	op.buf = buf;
	op.len = nbytes;
	op.fd = fd;

	if((hOutputThread = CreateThread(NULL, 0, WriteOp, &op, 0, &dwOutputThreadID)) == NULL)
	{
		vlog(VLOG_ERR, "ERROR: failed to create input operation thread.");
		return(-1);
	}

	/*
	 * Wait for the timeout period for the output operation to
	 * complete by itself, if the timeout expires, then abort
	 * the output operation in progress and return error status.
	 */

	switch(WaitForSingleObject(hOutputThread, (tout * 1000)))
	{
		case WAIT_OBJECT_0:	/* The Thread has exited by itself */
					break;
		case WAIT_TIMEOUT:	/* The timeout period expired */
					TerminateThread(hOutputThread, -1);
					break;
	}

	GetExitCodeThread(hOutputThread, &dwRetCode);

	return(dwRetCode);
}

DWORD WINAPI ReadOp(LPVOID arg)
{
	IO_OP	*op	= (IO_OP *)arg;
	DWORD	got	= 0;
	void	*buf	= op->buf;
	int	ret;

	/*
	 * Perform the input operation described by the IO_OP.
	 */

	while(got < op->len)
	{
		switch(ret = recv(op->fd, op->buf, (op->len - got), 0))
		{
			case SOCKET_ERROR:	ExitThread(-1);
			case 0:			ExitThread(-1);
			default:		got += ret;
						buf = (u_char *)buf + ret;
						break;
		}
	}

	return(got);
}
  
ssize_t vio_read(int fd, void *buf, size_t nbytes, time_t tout)
{
	HANDLE	hInputThread	= NULL;
	DWORD	dwInputThreadID	= 0;
	DWORD	dwRetCode	= -1;
	IO_OP	op;

	/*
	 * Create a Thread that will perform the requested input operation.
	 */

	op.buf = buf;
	op.len = nbytes;
	op.fd = fd;
	
	if((hInputThread = CreateThread(NULL, 0, ReadOp, &op, 0, &dwInputThreadID)) == NULL)
	{
		vlog(VLOG_ERR, "ERROR: failed to create input operation thread.");
		return(-1);
	}

	/*
	 * Wait for the timeout period for the input operation to
	 * complete by itself, if the timeout expires, then abort
	 * the input operation in progress and return error status.
	 */

	switch(WaitForSingleObject(hInputThread, (tout * 1000)))
	{
		case WAIT_OBJECT_0:	/* The Thread has exited by itself */
					break;
		case WAIT_TIMEOUT:	/* The timeout period expired */
					TerminateThread(hInputThread, -1);
					break;
	}

	GetExitCodeThread(hInputThread, &dwRetCode);

	return(dwRetCode);
}

#endif /* WIN32 */
